@@ -10,13 +10,14 @@
 use super::{new_qp2p, new_qp2p_with_hcc, random_msg};
 use anyhow::{anyhow, Result};
 use bytes::Bytes;
-use futures::future;
+use futures::{future, stream::FuturesUnordered, StreamExt};
 use std::{
     collections::{BTreeSet, HashSet},
     time::Duration,
 };
 use tiny_keccak::{Hasher, Sha3};
 use tokio::time::timeout;
+use tracing::info;
 use tracing_test::traced_test;
 /// SHA3-256 hash digest.
 type Digest256 = [u8; 32];
@@ -399,7 +400,7 @@ async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
             let mut sending_tasks = Vec::new();
 
             while let Some((src, msg)) = recv_incoming_messages.next().await {
-                tracing::info!("received from {:?} with message size {}", src, msg.len());
+                info!("received from {:?} with message size {}", src, msg.len());
                 assert_eq!(msg.len(), test_msgs[0].len());
 
                 let sending_endpoint = server_endpoint.clone();
@@ -442,29 +443,32 @@ async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
 
             async move {
                 let mut hash_results = BTreeSet::new();
-                tracing::info!("connecting {}", id);
                 send_endpoint.connect_to(&server_addr).await?;
                 for (index, message) in messages.iter().enumerate().take(num_messages_each) {
                     let _ = hash_results.insert(hash(&message));
-                    tracing::info!("sender #{} sending message #{}", id, index);
+                    info!("sender #{} sending message #{}", id, index);
                     send_endpoint
                         .send_message(message.clone(), &server_addr)
                         .await?;
                 }
 
-                tracing::info!(
+                info!(
                     "sender #{} completed sending messages, starts listening",
                     id
                 );
 
                 while let Some((src, msg)) = recv_incoming_messages.next().await {
-                    tracing::info!(
+                    info!(
                         "#{} received from server {:?} with message size {}",
                         id,
                         src,
                         msg.len()
                     );
+
+                    info!("Hash len before: {:?}", hash_results.len());
                     assert!(hash_results.remove(&msg[..]));
+                    info!("Hash len after: {:?}", hash_results.len());
+
                     if hash_results.is_empty() {
                         break;
                     }
@@ -479,6 +483,121 @@ async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
     Ok(())
 }
 
+#[tokio::test]
+#[traced_test]
+async fn multiple_connections_with_many_larger_concurrent_messages() -> Result<()> {
+    let num_senders: usize = 10;
+    let num_messages_each: usize = 100;
+    let num_messages_total: usize = num_senders * num_messages_each;
+
+    let qp2p = new_qp2p()?;
+    let (server_endpoint, _, mut recv_incoming_messages, _) = qp2p.new_endpoint().await?;
+    let server_addr = server_endpoint.socket_addr();
+
+    let test_msgs: Vec<_> = (0..num_messages_each)
+        .map(|_| random_msg(1024 * 1024))
+        .collect();
+    let sending_msgs = test_msgs.clone();
+
+    let mut tasks = FuturesUnordered::new();
+
+    // Receiver + Hasher
+    tasks.push(tokio::spawn({
+        async move {
+            let mut num_received = 0;
+            assert!(!logs_contain("error"));
+
+            while let Some((src, msg)) = recv_incoming_messages.next().await {
+                info!("received from {:?} with message size {}", src, msg.len());
+                assert!(!logs_contain("error"));
+
+                assert_eq!(msg.len(), test_msgs[0].len());
+
+                let sending_endpoint = server_endpoint.clone();
+
+                let hash_result = hash(&msg);
+                for _ in 0..5 {
+                    let _ = hash(&msg);
+                }
+
+                // Send the hash result back.
+                sending_endpoint
+                    .send_message(hash_result.to_vec().into(), &src)
+                    .await?;
+
+                assert!(!logs_contain("error"));
+
+                num_received += 1;
+                println!("Server received count: {}", num_received);
+                if num_received >= num_messages_total {
+                    break;
+                }
+            }
+
+            assert!(!logs_contain("error"));
+
+            Ok(())
+        }
+    }));
+
+    // Sender + Verifier
+    for id in 0..num_senders {
+        let messages = sending_msgs.clone();
+        assert!(!logs_contain("error"));
+
+        tasks.push(tokio::spawn({
+            let qp2p = new_qp2p()?;
+            let (send_endpoint, _, mut recv_incoming_messages, _) = qp2p.new_endpoint().await?;
+
+            async move {
+                let mut hash_results = BTreeSet::new();
+
+                send_endpoint.connect_to(&server_addr).await?;
+                for (index, message) in messages.iter().enumerate().take(num_messages_each) {
+                    let _ = hash_results.insert(hash(&message));
+
+                    info!("sender #{} sending message #{}", id, index);
+                    send_endpoint
+                        .send_message(message.clone(), &server_addr)
+                        .await?;
+                }
+
+                info!(
+                    "sender #{} completed sending messages, starts listening",
+                    id
+                );
+
+                while let Some((src, msg)) = recv_incoming_messages.next().await {
+                    assert!(!logs_contain("error"));
+
+                    info!(
+                        "sender #{} received from server {:?} with message size {:?}",
+                        id, src, msg
+                    );
+
+                    info!("Hash len before: {:?}", hash_results.len());
+                    assert!(hash_results.remove(&msg[..]));
+                    info!("Hash len after: {:?}", hash_results.len());
+
+                    if hash_results.is_empty() {
+                        break;
+                    }
+                }
+
+                Ok::<_, anyhow::Error>(())
+            }
+        }));
+    }
+
+    while let Some(result) = tasks.next().await {
+        match result {
+            Ok(Ok(())) => (),
+            other => anyhow::bail!("Error from test threads: {:?}", other??),
+        }
+    }
+    Ok(())
+}
+
 #[tokio::test]
 #[traced_test]
 async fn many_messages() -> Result<()> {
@@ -503,11 +622,11 @@ async fn many_messages() -> Result<()> {
         tasks.push(tokio::spawn({
             let endpoint = send_endpoint.clone();
             async move {
-                tracing::info!("sending {}", id);
+                info!("sending {}", id);
                 let msg = id.to_le_bytes().to_vec().into();
                 endpoint.connect_to(&recv_addr).await?;
                 endpoint.send_message(msg, &recv_addr).await?;
-                tracing::info!("sent {}", id);
+                info!("sent {}", id);
 
                 Ok::<_, anyhow::Error>(())
             }
@@ -522,7 +641,7 @@ async fn many_messages() -> Result<()> {
             while let Some((src, msg)) = recv_incoming_messages.next().await {
                 let id = usize::from_le_bytes(msg[..].try_into().unwrap());
                 assert_eq!(src, send_addr);
-                tracing::info!("received {}", id);
+                info!("received {}", id);
 
                 num_received += 1;
 
